==> bodo-2022.1-cp39-cp39-macosx_10_15_x86_64.whl: bodo/utils/indexing.py <==
"""
Collection of utility functions for indexing implementation (getitem/setitem)
"""
import operator
import numba
import numpy as np
from numba.core import types
from numba.core.imputils import impl_ret_borrowed
from numba.extending import intrinsic, overload, register_jitable
import bodo
from bodo.utils.typing import BodoError


@register_jitable
def get_new_null_mask_bool_index(old_mask, ind, n):
    kkq__nihmw = n + 7 >> 3
    bevw__awk = np.empty(kkq__nihmw, np.uint8)
    cnpr__fwdud = 0
    for lqu__mypfm in range(len(ind)):
        if ind[lqu__mypfm]:
            kiuhb__wojel = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
                lqu__mypfm)
            bodo.libs.int_arr_ext.set_bit_to_arr(bevw__awk, cnpr__fwdud,
                kiuhb__wojel)
            cnpr__fwdud += 1
    return bevw__awk


@register_jitable
def array_getitem_bool_index(A, ind):
    ind = bodo.utils.conversion.coerce_to_ndarray(ind)
    old_mask = A._null_bitmap
    lqb__xts = A._data[ind]
    n = len(lqb__xts)
    bevw__awk = get_new_null_mask_bool_index(old_mask, ind, n)
    return lqb__xts, bevw__awk


@register_jitable
def get_new_null_mask_int_index(old_mask, ind, n):
    kkq__nihmw = n + 7 >> 3
    bevw__awk = np.empty(kkq__nihmw, np.uint8)
    cnpr__fwdud = 0
    for lqu__mypfm in range(len(ind)):
        kiuhb__wojel = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
            ind[lqu__mypfm])
        bodo.libs.int_arr_ext.set_bit_to_arr(bevw__awk, cnpr__fwdud,
            kiuhb__wojel)
        cnpr__fwdud += 1
    return bevw__awk


@register_jitable
def array_getitem_int_index(A, ind):
    qkk__mix = bodo.utils.conversion.coerce_to_ndarray(ind)
    old_mask = A._null_bitmap
    lqb__xts = A._data[qkk__mix]
    n = len(lqb__xts)
    bevw__awk = get_new_null_mask_int_index(old_mask, qkk__mix, n)
    return lqb__xts, bevw__awk


@register_jitable
def get_new_null_mask_slice_index(old_mask, ind, n):
    hbp__mnwei = numba.cpython.unicode._normalize_slice(ind, n)
    auib__ubhs = numba.cpython.unicode._slice_span(hbp__mnwei)
    kkq__nihmw = auib__ubhs + 7 >> 3
    bevw__awk = np.empty(kkq__nihmw, np.uint8)
    cnpr__fwdud = 0
    for lqu__mypfm in range(hbp__mnwei.start, hbp__mnwei.stop, hbp__mnwei.step
        ):
        kiuhb__wojel = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
            lqu__mypfm)
        bodo.libs.int_arr_ext.set_bit_to_arr(bevw__awk, cnpr__fwdud,
            kiuhb__wojel)
        cnpr__fwdud += 1
    return bevw__awk


@register_jitable
def array_getitem_slice_index(A, ind):
    n = len(A._data)
    old_mask = A._null_bitmap
    lqb__xts = np.ascontiguousarray(A._data[ind])
    bevw__awk = get_new_null_mask_slice_index(old_mask, ind, n)
    return lqb__xts, bevw__awk


def array_setitem_int_index(A, idx, val):
    return


@overload(array_setitem_int_index, no_unliteral=True)
def array_setitem_int_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(val._data)
            for lqu__mypfm in range(n):
                A._data[idx[lqu__mypfm]] = val._data[lqu__mypfm]
                kiuhb__wojel = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val
                    ._null_bitmap, lqu__mypfm)
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap, idx[
                    lqu__mypfm], kiuhb__wojel)
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            for lqu__mypfm in idx:
                A._data[lqu__mypfm] = val
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                    lqu__mypfm, 1)
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


def array_setitem_bool_index(A, idx, val):
    A[idx] = val


@overload(array_setitem_bool_index, no_unliteral=True)
def array_setitem_bool_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(idx)
            noyg__pumm = 0
            for lqu__mypfm in range(n):
                if not bodo.libs.array_kernels.isna(idx, lqu__mypfm) and idx[
                    lqu__mypfm]:
                    A._data[lqu__mypfm] = val._data[noyg__pumm]
                    kiuhb__wojel = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val
                        ._null_bitmap, noyg__pumm)
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        lqu__mypfm, kiuhb__wojel)
                    noyg__pumm += 1
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            n = len(idx)
            noyg__pumm = 0
            for lqu__mypfm in range(n):
                if not bodo.libs.array_kernels.isna(idx, lqu__mypfm) and idx[
                    lqu__mypfm]:
                    A._data[lqu__mypfm] = val
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        lqu__mypfm, 1)
                    noyg__pumm += 1
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


@register_jitable
def setitem_slice_index_null_bits(dst_bitmap, src_bitmap, idx, n):
    hbp__mnwei = numba.cpython.unicode._normalize_slice(idx, n)
    noyg__pumm = 0
    for lqu__mypfm in range(hbp__mnwei.start, hbp__mnwei.stop, hbp__mnwei.step
        ):
        kiuhb__wojel = bodo.libs.int_arr_ext.get_bit_bitmap_arr(src_bitmap,
            noyg__pumm)
        bodo.libs.int_arr_ext.set_bit_to_arr(dst_bitmap, lqu__mypfm,
            kiuhb__wojel)
        noyg__pumm += 1


def array_setitem_slice_index(A, idx, val):
    return


@overload(array_setitem_slice_index, no_unliteral=True)
def array_setitem_slice_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(A._data)
            A._data[idx] = val._data
            src_bitmap = val._null_bitmap.copy()
            setitem_slice_index_null_bits(A._null_bitmap, src_bitmap, idx, n)
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            hbp__mnwei = numba.cpython.unicode._normalize_slice(idx, len(A))
            for lqu__mypfm in range(hbp__mnwei.start, hbp__mnwei.stop,
                hbp__mnwei.step):
                A._data[lqu__mypfm] = val
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                    lqu__mypfm, 1)
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


def untuple_if_one_tuple(v):
    return v


@overload(untuple_if_one_tuple)
def untuple_if_one_tuple_overload(v):
    if isinstance(v, types.BaseTuple) and len(v.types) == 1:
        return lambda v: v[0]
    return lambda v: v


def init_nested_counts(arr_typ):
    return 0,


@overload(init_nested_counts)
def overload_init_nested_counts(arr_typ):
    arr_typ = arr_typ.instance_type
    if isinstance(arr_typ, bodo.libs.array_item_arr_ext.ArrayItemArrayType
        ) or arr_typ == bodo.string_array_type:
        data_arr_typ = arr_typ.dtype
        return lambda arr_typ: (0,) + init_nested_counts(data_arr_typ)
    if bodo.utils.utils.is_array_typ(arr_typ, False
        ) or arr_typ == bodo.string_type:
        return lambda arr_typ: (0,)
    return lambda arr_typ: ()


def add_nested_counts(nested_counts, arr_item):
    return 0,


@overload(add_nested_counts)
def overload_add_nested_counts(nested_counts, arr_item):
    from bodo.libs.str_arr_ext import get_utf8_size
    arr_item = arr_item.type if isinstance(arr_item, types.Optional
        ) else arr_item
    if isinstance(arr_item, bodo.libs.array_item_arr_ext.ArrayItemArrayType):
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item),) + add_nested_counts(nested_counts[1:], bodo.libs.
            array_item_arr_ext.get_data(arr_item))
    if isinstance(arr_item, types.List):
        return lambda nested_counts, arr_item: add_nested_counts(nested_counts,
            bodo.utils.conversion.coerce_to_array(arr_item))
    if arr_item == bodo.string_array_type:
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item), nested_counts[1] + np.int64(bodo.libs.str_arr_ext.
            num_total_chars(arr_item)))
    if bodo.utils.utils.is_array_typ(arr_item, False):
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item),)
    if arr_item == bodo.string_type:
        return lambda nested_counts, arr_item: (nested_counts[0] +
            get_utf8_size(arr_item),)
    return lambda nested_counts, arr_item: ()


@overload(operator.setitem)
def none_optional_setitem_overload(A, idx, val):
    if not bodo.utils.utils.is_array_typ(A, False):
        return
    elif val == types.none:
        if isinstance(idx, types.Integer):
            return lambda A, idx, val: bodo.libs.array_kernels.setna(A, idx)
        elif bodo.utils.typing.is_list_like_index_type(idx) and isinstance(idx
            .dtype, types.Integer):

            def setitem_none_int_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                for lqu__mypfm in idx:
                    bodo.libs.array_kernels.setna(A, lqu__mypfm)
            return setitem_none_int_arr
        elif bodo.utils.typing.is_list_like_index_type(idx
            ) and idx.dtype == types.bool_:
            if A == bodo.string_array_type:

                def string_arr_impl(A, idx, val):
                    n = len(A)
                    idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                    hlko__gduiw = bodo.libs.str_arr_ext.pre_alloc_string_array(
                        n, -1)
                    for lqu__mypfm in numba.parfors.parfor.internal_prange(n):
                        if idx[lqu__mypfm] or bodo.libs.array_kernels.isna(A,
                            lqu__mypfm):
                            hlko__gduiw[lqu__mypfm] = ''
                            bodo.libs.str_arr_ext.str_arr_set_na(hlko__gduiw,
                                lqu__mypfm)
                        else:
                            hlko__gduiw[lqu__mypfm] = A[lqu__mypfm]
                    bodo.libs.str_arr_ext.move_str_binary_arr_payload(A,
                        hlko__gduiw)
                return string_arr_impl

            def setitem_none_bool_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                n = len(idx)
                for lqu__mypfm in range(n):
                    if not bodo.libs.array_kernels.isna(idx, lqu__mypfm
                        ) and idx[lqu__mypfm]:
                        bodo.libs.array_kernels.setna(A, lqu__mypfm)
            return setitem_none_bool_arr
        elif isinstance(idx, types.SliceType):

            def setitem_none_slice(A, idx, val):
                n = len(A)
                hbp__mnwei = numba.cpython.unicode._normalize_slice(idx, n)
                for lqu__mypfm in range(hbp__mnwei.start, hbp__mnwei.stop,
                    hbp__mnwei.step):
                    bodo.libs.array_kernels.setna(A, lqu__mypfm)
            return setitem_none_slice
        raise BodoError(
            f'setitem for {A} with indexing type {idx} and None value not supported.'
            )
    elif isinstance(val, types.optional):
        if isinstance(idx, types.Integer):

            def impl_optional(A, idx, val):
                if val is None:
                    bodo.libs.array_kernels.setna(A, idx)
                else:
                    A[idx] = bodo.utils.indexing.unoptional(val)
            return impl_optional
        elif bodo.utils.typing.is_list_like_index_type(idx) and isinstance(idx
            .dtype, types.Integer):

            def setitem_optional_int_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                for lqu__mypfm in idx:
                    if val is None:
                        bodo.libs.array_kernels.setna(A, lqu__mypfm)
                        continue
                    A[lqu__mypfm] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_int_arr
        elif bodo.utils.typing.is_list_like_index_type(idx
            ) and idx.dtype == types.bool_:
            if A == bodo.string_array_type:

                def string_arr_impl(A, idx, val):
                    if val is None:
                        A[idx] = None
                    else:
                        A[idx] = bodo.utils.indexing.unoptional(val)
                return string_arr_impl

            def setitem_optional_bool_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                n = len(idx)
                for lqu__mypfm in range(n):
                    if not bodo.libs.array_kernels.isna(idx, lqu__mypfm
                        ) and idx[lqu__mypfm]:
                        if val is None:
                            bodo.libs.array_kernels.setna(A, lqu__mypfm)
                            continue
                        A[lqu__mypfm] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_bool_arr
        elif isinstance(idx, types.SliceType):

            def setitem_optional_slice(A, idx, val):
                n = len(A)
                hbp__mnwei = numba.cpython.unicode._normalize_slice(idx, n)
                for lqu__mypfm in range(hbp__mnwei.start, hbp__mnwei.stop,
                    hbp__mnwei.step):
                    if val is None:
                        bodo.libs.array_kernels.setna(A, lqu__mypfm)
                        continue
                    A[lqu__mypfm] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_slice
        raise BodoError(
            f'setitem for {A} with indexing type {idx} and optional value not supported.'
            )


@intrinsic
def unoptional(typingctx, val_t=None):
    if not isinstance(val_t, types.Optional):
        return val_t(val_t), lambda c, b, s, args: impl_ret_borrowed(c, b,
            val_t, args[0])

    def codegen(context, builder, signature, args):
        oovn__msdi = context.make_helper(builder, val_t, args[0])
        zdz__qlvi = oovn__msdi.data
        context.nrt.incref(builder, val_t.type, zdz__qlvi)
        return zdz__qlvi
    return val_t.type(val_t), codegen

==> bodo-2022.2-cp39-cp39-macosx_10_15_x86_64.whl: bodo/utils/indexing.py <==
"""
Collection of utility functions for indexing implementation (getitem/setitem)
"""
import operator
import numba
import numpy as np
from numba.core import types
from numba.core.imputils import impl_ret_borrowed
from numba.extending import intrinsic, overload, register_jitable
import bodo
from bodo.utils.typing import BodoError


@register_jitable
def get_new_null_mask_bool_index(old_mask, ind, n):
    zzlt__wxves = n + 7 >> 3
    nxxmh__obgn = np.empty(zzlt__wxves, np.uint8)
    sqds__yixhd = 0
    for kztwb__xhn in range(len(ind)):
        if ind[kztwb__xhn]:
            pvbd__asy = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
                kztwb__xhn)
            bodo.libs.int_arr_ext.set_bit_to_arr(nxxmh__obgn, sqds__yixhd,
                pvbd__asy)
            sqds__yixhd += 1
    return nxxmh__obgn


@register_jitable
def array_getitem_bool_index(A, ind):
    ind = bodo.utils.conversion.coerce_to_ndarray(ind)
    old_mask = A._null_bitmap
    vsvcm__kxli = A._data[ind]
    n = len(vsvcm__kxli)
    nxxmh__obgn = get_new_null_mask_bool_index(old_mask, ind, n)
    return vsvcm__kxli, nxxmh__obgn


@register_jitable
def get_new_null_mask_int_index(old_mask, ind, n):
    zzlt__wxves = n + 7 >> 3
    nxxmh__obgn = np.empty(zzlt__wxves, np.uint8)
    sqds__yixhd = 0
    for kztwb__xhn in range(len(ind)):
        pvbd__asy = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask, ind[
            kztwb__xhn])
        bodo.libs.int_arr_ext.set_bit_to_arr(nxxmh__obgn, sqds__yixhd,
            pvbd__asy)
        sqds__yixhd += 1
    return nxxmh__obgn


@register_jitable
def array_getitem_int_index(A, ind):
    fyb__keo = bodo.utils.conversion.coerce_to_ndarray(ind)
    old_mask = A._null_bitmap
    vsvcm__kxli = A._data[fyb__keo]
    n = len(vsvcm__kxli)
    nxxmh__obgn = get_new_null_mask_int_index(old_mask, fyb__keo, n)
    return vsvcm__kxli, nxxmh__obgn


@register_jitable
def get_new_null_mask_slice_index(old_mask, ind, n):
    rbgf__jkz = numba.cpython.unicode._normalize_slice(ind, n)
    uusdf__xykzs = numba.cpython.unicode._slice_span(rbgf__jkz)
    zzlt__wxves = uusdf__xykzs + 7 >> 3
    nxxmh__obgn = np.empty(zzlt__wxves, np.uint8)
    sqds__yixhd = 0
    for kztwb__xhn in range(rbgf__jkz.start, rbgf__jkz.stop, rbgf__jkz.step):
        pvbd__asy = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
            kztwb__xhn)
        bodo.libs.int_arr_ext.set_bit_to_arr(nxxmh__obgn, sqds__yixhd,
            pvbd__asy)
        sqds__yixhd += 1
    return nxxmh__obgn


@register_jitable
def array_getitem_slice_index(A, ind):
    n = len(A._data)
    old_mask = A._null_bitmap
    vsvcm__kxli = np.ascontiguousarray(A._data[ind])
    nxxmh__obgn = get_new_null_mask_slice_index(old_mask, ind, n)
    return vsvcm__kxli, nxxmh__obgn


def array_setitem_int_index(A, idx, val):
    return


@overload(array_setitem_int_index, no_unliteral=True)
def array_setitem_int_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(val._data)
            for kztwb__xhn in range(n):
                A._data[idx[kztwb__xhn]] = val._data[kztwb__xhn]
                pvbd__asy = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val.
                    _null_bitmap, kztwb__xhn)
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap, idx[
                    kztwb__xhn], pvbd__asy)
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            for kztwb__xhn in idx:
                A._data[kztwb__xhn] = val
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                    kztwb__xhn, 1)
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


def array_setitem_bool_index(A, idx, val):
    A[idx] = val


@overload(array_setitem_bool_index, no_unliteral=True)
def array_setitem_bool_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(idx)
            ankux__ylrde = 0
            for kztwb__xhn in range(n):
                if not bodo.libs.array_kernels.isna(idx, kztwb__xhn) and idx[
                    kztwb__xhn]:
                    A._data[kztwb__xhn] = val._data[ankux__ylrde]
                    pvbd__asy = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val
                        ._null_bitmap, ankux__ylrde)
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        kztwb__xhn, pvbd__asy)
                    ankux__ylrde += 1
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            n = len(idx)
            ankux__ylrde = 0
            for kztwb__xhn in range(n):
                if not bodo.libs.array_kernels.isna(idx, kztwb__xhn) and idx[
                    kztwb__xhn]:
                    A._data[kztwb__xhn] = val
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        kztwb__xhn, 1)
                    ankux__ylrde += 1
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


@register_jitable
def setitem_slice_index_null_bits(dst_bitmap, src_bitmap, idx, n):
    rbgf__jkz = numba.cpython.unicode._normalize_slice(idx, n)
    ankux__ylrde = 0
    for kztwb__xhn in range(rbgf__jkz.start, rbgf__jkz.stop, rbgf__jkz.step):
        pvbd__asy = bodo.libs.int_arr_ext.get_bit_bitmap_arr(src_bitmap,
            ankux__ylrde)
        bodo.libs.int_arr_ext.set_bit_to_arr(dst_bitmap, kztwb__xhn, pvbd__asy)
        ankux__ylrde += 1


def array_setitem_slice_index(A, idx, val):
    return


@overload(array_setitem_slice_index, no_unliteral=True)
def array_setitem_slice_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(A._data)
            A._data[idx] = val._data
            src_bitmap = val._null_bitmap.copy()
            setitem_slice_index_null_bits(A._null_bitmap, src_bitmap, idx, n)
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            rbgf__jkz = numba.cpython.unicode._normalize_slice(idx, len(A))
            for kztwb__xhn in range(rbgf__jkz.start, rbgf__jkz.stop,
                rbgf__jkz.step):
                A._data[kztwb__xhn] = val
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                    kztwb__xhn, 1)
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


def untuple_if_one_tuple(v):
    return v


@overload(untuple_if_one_tuple)
def untuple_if_one_tuple_overload(v):
    if isinstance(v, types.BaseTuple) and len(v.types) == 1:
        return lambda v: v[0]
    return lambda v: v


def init_nested_counts(arr_typ):
    return 0,


@overload(init_nested_counts)
def overload_init_nested_counts(arr_typ):
    arr_typ = arr_typ.instance_type
    if isinstance(arr_typ, bodo.libs.array_item_arr_ext.ArrayItemArrayType
        ) or arr_typ == bodo.string_array_type:
        data_arr_typ = arr_typ.dtype
        return lambda arr_typ: (0,) + init_nested_counts(data_arr_typ)
    if bodo.utils.utils.is_array_typ(arr_typ, False
        ) or arr_typ == bodo.string_type:
        return lambda arr_typ: (0,)
    return lambda arr_typ: ()


def add_nested_counts(nested_counts, arr_item):
    return 0,


@overload(add_nested_counts)
def overload_add_nested_counts(nested_counts, arr_item):
    from bodo.libs.str_arr_ext import get_utf8_size
    arr_item = arr_item.type if isinstance(arr_item, types.Optional
        ) else arr_item
    if isinstance(arr_item, bodo.libs.array_item_arr_ext.ArrayItemArrayType):
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item),) + add_nested_counts(nested_counts[1:], bodo.libs.
            array_item_arr_ext.get_data(arr_item))
    if isinstance(arr_item, types.List):
        return lambda nested_counts, arr_item: add_nested_counts(nested_counts,
            bodo.utils.conversion.coerce_to_array(arr_item))
    if arr_item == bodo.string_array_type:
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item), nested_counts[1] + np.int64(bodo.libs.str_arr_ext.
            num_total_chars(arr_item)))
    if bodo.utils.utils.is_array_typ(arr_item, False):
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item),)
    if arr_item == bodo.string_type:
        return lambda nested_counts, arr_item: (nested_counts[0] +
            get_utf8_size(arr_item),)
    return lambda nested_counts, arr_item: ()


@overload(operator.setitem)
def none_optional_setitem_overload(A, idx, val):
    if not bodo.utils.utils.is_array_typ(A, False):
        return
    elif val == types.none:
        if isinstance(idx, types.Integer):
            return lambda A, idx, val: bodo.libs.array_kernels.setna(A, idx)
        elif bodo.utils.typing.is_list_like_index_type(idx) and isinstance(idx
            .dtype, types.Integer):

            def setitem_none_int_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                for kztwb__xhn in idx:
                    bodo.libs.array_kernels.setna(A, kztwb__xhn)
            return setitem_none_int_arr
        elif bodo.utils.typing.is_list_like_index_type(idx
            ) and idx.dtype == types.bool_:
            if A == bodo.string_array_type:

                def string_arr_impl(A, idx, val):
                    n = len(A)
                    idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                    lcnu__xxd = bodo.libs.str_arr_ext.pre_alloc_string_array(n,
                        -1)
                    for kztwb__xhn in numba.parfors.parfor.internal_prange(n):
                        if idx[kztwb__xhn] or bodo.libs.array_kernels.isna(A,
                            kztwb__xhn):
                            lcnu__xxd[kztwb__xhn] = ''
                            bodo.libs.str_arr_ext.str_arr_set_na(lcnu__xxd,
                                kztwb__xhn)
                        else:
                            lcnu__xxd[kztwb__xhn] = A[kztwb__xhn]
                    bodo.libs.str_arr_ext.move_str_binary_arr_payload(A,
                        lcnu__xxd)
                return string_arr_impl

            def setitem_none_bool_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                n = len(idx)
                for kztwb__xhn in range(n):
                    if not bodo.libs.array_kernels.isna(idx, kztwb__xhn
                        ) and idx[kztwb__xhn]:
                        bodo.libs.array_kernels.setna(A, kztwb__xhn)
            return setitem_none_bool_arr
        elif isinstance(idx, types.SliceType):

            def setitem_none_slice(A, idx, val):
                n = len(A)
                rbgf__jkz = numba.cpython.unicode._normalize_slice(idx, n)
                for kztwb__xhn in range(rbgf__jkz.start, rbgf__jkz.stop,
                    rbgf__jkz.step):
                    bodo.libs.array_kernels.setna(A, kztwb__xhn)
            return setitem_none_slice
        raise BodoError(
            f'setitem for {A} with indexing type {idx} and None value not supported.'
            )
    elif isinstance(val, types.optional):
        if isinstance(idx, types.Integer):

            def impl_optional(A, idx, val):
                if val is None:
                    bodo.libs.array_kernels.setna(A, idx)
                else:
                    A[idx] = bodo.utils.indexing.unoptional(val)
            return impl_optional
        elif bodo.utils.typing.is_list_like_index_type(idx) and isinstance(idx
            .dtype, types.Integer):

            def setitem_optional_int_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                for kztwb__xhn in idx:
                    if val is None:
                        bodo.libs.array_kernels.setna(A, kztwb__xhn)
                        continue
                    A[kztwb__xhn] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_int_arr
        elif bodo.utils.typing.is_list_like_index_type(idx
            ) and idx.dtype == types.bool_:
            if A == bodo.string_array_type:

                def string_arr_impl(A, idx, val):
                    if val is None:
                        A[idx] = None
                    else:
                        A[idx] = bodo.utils.indexing.unoptional(val)
                return string_arr_impl

            def setitem_optional_bool_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                n = len(idx)
                for kztwb__xhn in range(n):
                    if not bodo.libs.array_kernels.isna(idx, kztwb__xhn
                        ) and idx[kztwb__xhn]:
                        if val is None:
                            bodo.libs.array_kernels.setna(A, kztwb__xhn)
                            continue
                        A[kztwb__xhn] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_bool_arr
        elif isinstance(idx, types.SliceType):

            def setitem_optional_slice(A, idx, val):
                n = len(A)
                rbgf__jkz = numba.cpython.unicode._normalize_slice(idx, n)
                for kztwb__xhn in range(rbgf__jkz.start, rbgf__jkz.stop,
                    rbgf__jkz.step):
                    if val is None:
                        bodo.libs.array_kernels.setna(A, kztwb__xhn)
                        continue
                    A[kztwb__xhn] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_slice
        raise BodoError(
            f'setitem for {A} with indexing type {idx} and optional value not supported.'
            )


@intrinsic
def unoptional(typingctx, val_t=None):
    if not isinstance(val_t, types.Optional):
        return val_t(val_t), lambda c, b, s, args: impl_ret_borrowed(c, b,
            val_t, args[0])

    def codegen(context, builder, signature, args):
        giihy__dbd = context.make_helper(builder, val_t, args[0])
        gxs__qao = giihy__dbd.data
        context.nrt.incref(builder, val_t.type, gxs__qao)
        return gxs__qao
    return val_t.type(val_t), codegen
